#ifndef IMAPFILTER_H
#define IMAPFILTER_H


#include <stdio.h>
#include <sys/types.h>

#include "config.h"
#include "account.h"
#include "filter.h"

#ifdef SSL_TLS
#include <openssl/ssl.h>
#endif


/* Error codes returned by functions. */
#define ERROR_SIGNAL			1
#define ERROR_TRIVIAL			2
#define ERROR_FILE_OPEN			3
#define ERROR_LOCKFILE			4
#define ERROR_CONFIG_PARSE		5
#define ERROR_MEMORY_ALLOCATION		6
#define ERROR_SETUID			7
#define ERROR_TERMIO			8
#define ERROR_NETWORK			9
#define ERROR_SSL			10
#define ERROR_PASSPHRASE		11
#define ERROR_ENCRYPT			12
#define ERROR_DECRYPT			13
#define ERROR_FORK			14
#define ERROR_UNDEFINED			15

/* SSL/TLS certificate status for the server. */
#define SSL_CERT_OK			0
#define SSL_CERT_NONEXISTENT		1
#define SSL_CERT_MISMATCH		2

/* Action to be applied, concerning the SSL/TLS certificate. */
#define SSL_CERT_ACTION_ACCEPT		0
#define SSL_CERT_ACTION_REJECT		1
#define SSL_CERT_ACTION_CONTINUE	0
#define SSL_CERT_ACTION_ABORT		1

/* Flags that control the program's execution options. */
#define OPTION_DETAILS_QUIET		0x0001
#define OPTION_DETAILS_NORMAL		0x0002
#define OPTION_DETAILS_VERBOSE		0x0004
#define OPTION_DETAILS_CLEAR		0xfff8
#define OPTION_ERRORS			0x0008
#define OPTION_EXPUNGE			0x0010
#define OPTION_HEADERS			0x0020
#define OPTION_NAMESPACE		0x0040
#define OPTION_SUBSCRIBE		0x0080
#define OPTION_PASSWORD_EDITOR		0x0100
#define OPTION_DAEMON_MODE		0x0200

/* Other flags. */
#define FLAG_DAEMON_MODE		0x01
#define FLAG_TTY			0x02
#define FLAG_TTY_MODIFIED		0x04
#define FLAG_BLANK_PASSWORD		0x08
#define FLAG_SIGUSR1_RECEIVED		0x10

/* Capabilities of mail server. */
#define CAPABILITY_NONE                 0x00
#define CAPABILITY_NAMESPACE            0x01
#define CAPABILITY_AUTH_CRAM_MD5	0x02
#define CAPABILITY_STARTTLS		0x04

/* Flags for logger. */
#define LOG_PREAMBLE			0
#define LOG_ACCOUNT			1
#define LOG_MAILBOX			2
#define LOG_FILTER			3
#define LOG_ACTION			4
#define LOG_DESTINATION_ACCOUNT		5
#define LOG_DESTINATION_MAILBOX		6
#define LOG_HEADER			7

/* Store IMAP command type may replace,add or remove flags. */
#define STORE_FLAG_REPLACE		1
#define STORE_FLAG_ADD			2
#define STORE_FLAG_REMOVE		3

/* Buffer size of server's response. */
#define RESPONSE_BUF			4096
#define RESULT_BUF			512

/* Response codes. */
#define RESPONSE_NULLBODY		-2
#define RESPONSE_NONE			-1
#define RESPONSE_OK			0
#define RESPONSE_NO			1
#define RESPONSE_BAD			2
#define RESPONSE_BYE			3
#define RESPONSE_PREAUTH		4
#define RESPONSE_READONLY		5
#define RESPONSE_TRYCREATE		6

/* Encryption and decryption buffers. */
#define ENCRYPTION_BUF			1024
#define DECRYPTION_BUF			4096

/* Buffer to save flags of the store IMAP command. */
#define STORE_FLAGS_BUF			64

/* Lenght of charset for IMAP SEARCH requests. */
#define CHARSET_LEN			64

/* String lengths of master password and encrypted passwords. */
#define PASSPHRASE_LEN			256

/* Length of mailbox namespace prefix. */
#define NAMESPACE_PREFIX_LEN            64


#define min(A, B)			((A) < (B) ? (A) : (B))
#define plural(A)			((A) == 1 ? "" : "s")


/* IMAP connection. */
typedef struct conn {
	int sock;		/* Socket. */
#ifdef SSL_TLS
	SSL *ssl;		/* SSL socket. */
#endif
	unsigned int caps;	/* Capabilities of the mail server. */
	struct {		/* Namespace of the mail server's mailboxes. */
		char prefix[NAMESPACE_PREFIX_LEN];	/* Namespace prefix. */
		char delim;	/* Namespace delimiter. */
	} nsp;
} conn_t;


/*	action.c	*/
int apply_action(char *mbox, char *mesgs, unsigned int *type, account_t * raccount, char *destmbox, unsigned int *msgflags, char *args);

/*	auth.c		*/
#ifdef CRAM_MD5
int auth_cram_md5(conn_t * conn, char *user, char *pass);
#endif

/*	cert.c		*/
#ifdef SSL_TLS
int get_cert(conn_t * conn);
#endif

/*	destroy.c	*/
void destroy_all(void);
void destroy_unneeded(void);

/*	file.c		*/
int create_homedir(void);
int exists_file(char *fname);
int exists_dir(char *fname);
int create_file(char *fname, mode_t mode);
#ifdef CHECK_PERMISSIONS
int check_file_perms(char *fname, mode_t mode);
int check_dir_perms(char *dname, mode_t mode);
#endif

/*	imap.c		*/
#ifdef DEBUG
int imap_noop(conn_t * conn);
#endif
int imap_capability(conn_t * conn);
int imap_namespace(conn_t * conn);
int imap_logout(conn_t * conn);
#ifdef SSL_TLS
int imap_starttls(conn_t * conn);
#endif
int imap_authenticate(conn_t * conn, char *auth, int cont);
int imap_login(conn_t * conn, char *user, char *pass);
/* int imap_list(conn_t *conn, char *refer, char *mbox); */
int imap_subscribe(conn_t * conn, char *mbox);
/* int imap_examine(conn_t *conn, char *mbox); */
int imap_select(conn_t * conn, char *mbox);
int imap_status(conn_t * conn, char *mbox, char *items);
int imap_create(conn_t * conn, char *mbox);
int imap_search(conn_t * conn, char *charset, char *search);
int imap_fetch(conn_t * conn, char *mesg, char *items);
int imap_store(conn_t * conn, char *mesg, unsigned int mode, char *flags);
int imap_copy(conn_t * conn, char *mesg, char *mbox);
int imap_append(conn_t * conn, char *mbox, unsigned int size);
int imap_close(conn_t * conn);
int imap_expunge(conn_t * conn);

/*	lock.c		*/
void lockfile_create(void);
void lockfile_check(void);
int lockfile_remove(void);
void kill_imapfilter(void);

/*	log.c		*/
void info(const char *info,...);
void verbose(const char *info,...);
void error(const char *errmsg,...);
void fatal(unsigned int errnum, const char *fatal,...);
void catch_signals(void);
int open_logfile(void);
int close_logfile(void);
void log_info(int flag, void *ptr);

/*	match.c		*/
int apply_filters(char *mbox, filter_t ** filters);

/*	memory.c	*/
void *xmalloc(size_t size);
void *xrealloc(void *ptr, size_t size);
void xfree(void *ptr);
char *xstrdup(const char *s);

void *smalloc(size_t size);
void *srealloc(void *ptr, size_t size);
void sfree(void *ptr);
char *sstrdup(const char *s);
void secmem_clear(void);

void corefile_disable(void);

/*	misc.c		*/
char *strcasestr(const char *haystack, const char *needle);
char *ultostr(unsigned long int num, int base);
char *xstrncpy(char *dest, const char *src, size_t size);

/*	parse.c		*/
int read_config(char *cfg);
void reread_config(char *cfg);
int read_passwords(void);

/*	passwd.c	*/
void get_password(char *passwd, size_t pwlen);
#ifdef ENCRYPTED_PASSWORDS
int encrypt_passwords(FILE * fd, account_t ** accts);
int decrypt_passwords(unsigned char **buf, FILE * fd);
void password_editor(void);
#endif

/*	request.c	*/
int test(conn_t * conn);
int check_capabilities(conn_t * conn);
int check_namespace(conn_t * conn);
#ifdef SSL_TLS
int negotiate_tls(conn_t * conn);
#endif
int login(conn_t * conn, char *user, char *pass);
int check_mailbox(conn_t * conn, char *mbox);
int select_mailbox(conn_t * conn, char *mbox);
int mailbox_status(conn_t * conn, char *mbox);
int close_mailbox(conn_t * conn);
int logout(conn_t * conn);

/*	response.c	*/
int server_response(conn_t * conn, unsigned int tag);
int greeting_response(conn_t * conn);
int logout_response(conn_t * conn, unsigned int tag);
int capability_response(conn_t * conn, unsigned int tag);
int authenticate_response(conn_t * conn, unsigned int tag, unsigned char **cont);
int namespace_response(conn_t * conn, unsigned int tag);
int status_response(conn_t * conn, unsigned int tag, char *mbox);
int select_response(conn_t * conn, unsigned int tag);
int search_response(conn_t * conn, unsigned int tag, char **mesgs);
int fetch_response(conn_t * conn, unsigned int tag, int reset, char *fetch);
int fetchsize_response(conn_t * conn, unsigned int *size, unsigned int tag);
int append_response(conn_t * conn, unsigned int tag);
int copy_response(conn_t * conn, unsigned int tag);

/*	socket.c	*/
int init_connection(conn_t * conn, char *serv, unsigned short int port, unsigned int protocol);
#ifdef SSL_TLS
int init_secure_connection(conn_t * conn, unsigned int protocol);
#endif
int close_connection(conn_t * conn);
int socket_read(conn_t * conn, char *buf);
int socket_write(conn_t * conn, char *data);

/*	tty.c		*/
int tty_store(void);
int tty_disable_echo(void);
int tty_restore(void);


#endif				/* IMAPFILTER_H */
